{-# LANGUAGE TemplateHaskell,
             FlexibleInstances,
             FlexibleContexts,
             MultiParamTypeClasses,
             ScopedTypeVariables,
             TypeOperators,
             DeriveDataTypeable,
             RankNTypes
 #-}

module AOP.Examples.Fib (fib, makeAspectFib)
where

import AOP.Default
import Unsafe.Coerce

fibTag = newTag

fibBase :: Monad m => Function Integer (m Integer)

fibPc :: (Typeable1Monad m, Typeable1Monad (t m)) => 
         NIPC Integer (NIAOT t m Integer) t

fibAdv :: NIAdvice Integer Integer t

fib :: (Typeable1Monad m, Typeable1Monad (t m), MonadError String (t m)) => 
       NIAOT t m (Integer -> NIAOT t m Integer)

makeAspectFib :: (Typeable1Monad m, Typeable1Monad (t m), 
                  LessGen (Integer -> NIAOT t m Integer) (a -> NIAOT t m b)) => 
                 NIAdvice a b t -> 
                 Aspect (NIAOT t m) (Integer -> NIAOT t m Integer) (a -> NIAOT t m b)


fibBase = mkFunction (\n_ -> return 1) fibTag

pcArgGT :: (Monad m, Ord a) => a -> RequirePC m (a -> b)
pcArgGT n = RequirePC (return (\jp -> return ((unsafeCoerce (getJpArg jp)) >= n)))

fibPc = pcCall fibBase `pcAnd` pcArgGT 2

fibAdv proceed n = do f1 <- fibBase # (n-1)
                      f2 <- fibBase # (n-2)
                      return (f1 + f2)
                   
fib = do deploy (niAspect fibPc fibAdv)
         return errorFib 
      where errorFib n = if n < 0 
                         then throwError "Not defined on negative value"
                         else fibBase # n

makeAspectFib adv = niAspect fibPc adv

runMemo c = runIdentity (runErrorT (runNIAOT c))

program n = runMemo $ do f <- fib
                         f # n
