{-# LANGUAGE FlexibleContexts #-}

module MemoizedFib (fib) where
import qualified AOP.Examples.Fib as Fib
import AOP.Default
import Data.Map
import Control.Monad.Error

memo :: Ord a => NIAdvice a b (StateT (Map a b))

memo proceed n =
     do m <- get
        if member n m
        then return (m ! n)
        else do y <- proceed n
                m' <- get
                put (insert n y m')
                return y

fib = do deploy (Fib.makeAspectFib memo)
         Fib.fib

runMemo c = runIdentity (runErrorT (evalStateT (runNIAOT c) empty))

program n = runMemo $ do f <- fib
                         f # n


--------------------------------------------------------------------------------
-- Memoization as narrowing advice

memo' :: (MonadState (Map a b) m, Ord a) => Narrow a b () m
memo' = (p, (bef, aft), rep) where
     p x       = do {m <- get; return (not (member x m))}
     bef _     = return ()
     aft x r _ = do {m <- get; put (insert x r m)}
     rep x     = do {m <- get; return (m ! x)}

fib' = do deploy (Fib.makeAspectFib (narrow memo'))
          Fib.fib

program' n = runMemo $ do f <- fib'
                          f # n
