{-# LANGUAGE ExistentialQuantification,
             ScopedTypeVariables,
             FlexibleContexts,
             DeriveDataTypeable,
             RankNTypes,
             GADTs
 #-}

module AOP.Internal.AspectDef ( 
 aspect,
 purePcAspect,
 pureAdviceAspect,
 pureAspect,
 newAspectHandle
) where

import AOP.Internal.Pointcut
import AOP.Internal.Advice
import AOP.Internal.LessGen
import AOP.Internal.Aspect
import AOP.Internal.Typeable1Monad
import Data.Unique
import System.IO.Unsafe

-- | Creates a new AspectHandle, i.e. a unique aspect identifier. 
newAspectHandle :: AspectHandle
newAspectHandle = unsafePerformIO newUnique

aspect :: (Monad m, LessGen t1 t2) => PC m t1 -> Advice t2 -> Aspect m t1 t2
aspect pc adv = Aspect pc adv newAspectHandle

-- | Pure PC/Advice. Purity is obtained by parameterizing the definitions on every monad m.
type PurePC a b = forall m. Typeable1Monad m => PC m (a -> b) 
type PureAdvice a b = forall m. Typeable1Monad m => Advice (a -> m b) 

-- | Builds an aspect with a pure pointcut
purePcAspect :: (Typeable1Monad m, LessGen (a -> m b) (c -> m d)) => 
                PurePC a (m b) -> Advice (c -> m d) -> Aspect m (a -> m b) (c -> m d)
purePcAspect pc adv = aspect pc adv

-- | Builds an aspect with a pure advice
pureAdviceAspect :: (Typeable1Monad m, LessGen (a -> m b) (c -> m d)) => 
                    PC m (a -> m b) -> PureAdvice c d -> Aspect m (a -> m b) (c -> m d)
pureAdviceAspect pc adv = aspect pc adv

-- | Builds an aspect with pure pointcut and advice
pureAspect :: (Typeable1Monad m, LessGen (a -> m b) (c -> m d)) => 
              PurePC a (m b) -> PureAdvice c d -> Aspect m (a -> m b) (c -> m d)
pureAspect pc adv = aspect pc adv