{-# LANGUAGE DeriveDataTypeable #-}

module AOP.Internal.Function (
 Function(..),
 FunctionTag,
 mkFunction,
 defaultFunctionTag,
 newTag
) where

import AOP.Internal.Typeable1Monad
import System.IO.Unsafe
import Data.Unique

{- |
We define a notion of function equality based on tagging functions with identifiers: two functions are equal if they have the same identifier.
We extend join point definition and weaving to support these tags. To provide compatibility with regular functions (which use StableNames for equality)
we use a special default tag in join points and weaving, to distinguish regular functions from tagged functions and use the corresponding notion of equality.
-}


type FunctionTag = Unique

data Function a b = Function {fun :: a -> b, tag :: FunctionTag} deriving Typeable

-- | Creates a tag to be used for regular functions. GHC guarantees that newUnique yields a unique value every time it is applied.
defaultFunctionTag = unsafePerformIO newUnique

-- | Creates a tag to be used for tagged function. GHC guarantees that newUnique yields a unique value every time it is applied.
newTag = unsafePerformIO newUnique

-- | Wraps a regular function into a wrapped one.
mkFunction :: (a -> b) -> FunctionTag -> Function a b
mkFunction f t = Function f t

-- | Equality of wrapped function is based on tag comparison. By construction tags are already instances of Eq.
instance Eq (Function a b) where
         Function _ tag == Function _ tag' = tag == tag'

