{-# LANGUAGE ExistentialQuantification,
             ScopedTypeVariables,
             FlexibleContexts,
             DeriveDataTypeable,
             RankNTypes,
             GADTs
 #-}

module AOP.Internal.NonInterferenceDef ( 
 niPcAspect,
 niAdviceAspect,
 niAspect,
 niBase,
 NIPC,
 NIAdvice
) where

import AOP.Internal.Pointcut
import AOP.Internal.Advice
import AOP.Internal.LessGen
import AOP.Internal.PolyTypeable
import AOP.Internal.Aspect
import AOP.Internal.AOT
import AOP.Internal.NIAOT
import AOP.Internal.AOPMonad
import AOP.Internal.Typeable1Monad
import AOP.Internal.AspectDef
import Data.Typeable
import Data.Unique
import System.IO.Unsafe

-- | Types of Non-Interfering poincut, advice, and base code
-- | The non-interference property is obtained by parameterizing on the monad m, or the monad transformer t.
type NIPC a b t = forall m. (Typeable1Monad m, Typeable1Monad (t m)) => PC (NIAOT t m) (a -> b) 

type NIAdvice a b t = forall m. (Typeable1Monad m, Typeable1Monad (t m)) => Advice (a -> NIAOT t m b) 

type NIBase a b m = forall t. (MonadTrans t, Typeable1Monad (t m)) => a -> NIAOT t m b

-- | Builds an aspect with a NI pointcut
niPcAspect :: (Typeable1Monad m, Typeable1Monad (t m), LessGen (a -> NIAOT t m b) (c -> NIAOT t m d)) => 
           NIPC a (NIAOT t m b) t -> Advice (c -> NIAOT t m d) -> 
           Aspect (NIAOT t m) (a -> NIAOT t m b) (c -> NIAOT t m d)
niPcAspect pc adv = aspect pc adv

-- | Builds an aspect with a NI advice
niAdviceAspect :: (Typeable1Monad m, Typeable1Monad (t m), LessGen (a -> NIAOT t m b) (c -> NIAOT t m d)) => 
            PC (NIAOT t m) (a -> NIAOT t m b) -> NIAdvice c d t ->
            Aspect (NIAOT t m) (a -> NIAOT t m b) (c -> NIAOT t m d)
niAdviceAspect pc adv = aspect pc adv

-- | Builds a NI aspect, that is it has NI pointcut and advice
niAspect :: (Typeable1Monad m, Typeable1Monad (t m), LessGen (a -> NIAOT t m b) (c -> NIAOT t m d)) => 
           NIPC a (NIAOT t m b) t -> NIAdvice c d t -> 
           Aspect (NIAOT t m) (a -> NIAOT t m b) (c -> NIAOT t m d)
niAspect pc adv = aspect pc adv

-- | Builds a NI base program
niBase :: (MonadTrans t, Typeable1Monad (t m)) => NIBase a b m -> a -> NIAOT t m b  
niBase f = f
