{-# LANGUAGE ExistentialQuantification,
             ScopedTypeVariables,
             RankNTypes
 #-}
module AOP.Internal.Pointcut (
 PC (..),
 RequirePC (..),
 runPC,
) where

import AOP.Internal.Joinpoint
import Data.Typeable

-- | A pointcut is a predicate on the current join point. It is used to identify join points of interest.
data PC m t = Monad m => PC {mpcond :: forall a b. m (Jp a b -> m Bool)} 

-- | Extracts the computation resulting of applying a join point to the pointcut
runPC :: Monad m => (PC m t) -> Jp a b -> m Bool
runPC (PC mpcond) jp = do pccond <- mpcond
                          pccond jp

-- | A RequirePC is not a valid standalone pointcut, it reflects a type requirement and must be combined with a standard PC.
data RequirePC m t = Monad m => RequirePC {mpcond' ::  forall a b. m (Jp a b -> m Bool)}

-- | Support for PolyTypeable
instance (Typeable1 m) => Typeable1 (PC m) where
         typeOf1 _ = mkTyConApp (mkTyCon3 "PC" "PC" "PC") 
                     [typeOf1 (undefined :: m ())]

-- | Support for PolyTypeable
instance (Typeable1 m) => Typeable1 (RequirePC m) where
         typeOf1 _ = mkTyConApp (mkTyCon3 "RequirePC" "RequirePC" "RequirePC") 
                     [typeOf1 (undefined :: m ())]



