{-# LANGUAGE FlexibleInstances,
             MultiParamTypeClasses,
             UndecidableInstances,
             ScopedTypeVariables,
             FlexibleContexts
 #-}

module AOP.Semantics.EL.ELT (
 ELT (ELT),
 Level,
 MonadEL(..),
 runELT,
) where

import AOP.Internal.Typeable1Monad
import AOP.Internal.PolyTypeableUtils
import AOP.Internal.Pointcut
import AOP.Internal.Aspect
import AOP.Internal.Advice
import AOP.Internal.AOPMonad
import AOP.Internal.AOT
import AOP.Internal.AOTType hiding (run)
import qualified AOP.Internal.AOTType as AOT (run)
import AOP.Semantics.EL.MonadEL

import Unsafe.Coerce

newtype ELT m a = ELT { run :: Level -> m (a, Level) }

runELT :: Monad m => ELT m a -> Level -> m a
runELT c l = do
               (v, _) <- run c l
               return v

inc :: Monad m => ELT m ()
inc = ELT (\l -> return ((), l+1))

dec :: Monad m => ELT m ()
dec = ELT (\l -> return ((), max (l - 1) 0))

at :: Monad m => Level -> ELT m ()
at l = ELT (\_ -> return ((), l))

-- | Monadic bind and return functions.
instance (Monad m) => Monad (ELT m) where
    return k = ELT $ \l -> return (k, l)
    ma >>= f = ELT $ \l0 -> do
                               (value, l1) <- run ma l0
                               run (f value) l1

-- | ELT produces level-aware functions.
instance (Monad m) => MonadEL (ELT m) where
  current = ELT (\l -> return (l, l))  
  up c   = do {inc; result <- c; dec; return result}
  down c = do {dec; result <- c; inc; return result}
  lambda_at f l arg = do n <- current
                         at l
                         result <- f arg
                         at n
                         return result

{- Interaction with the standard monad transformers, and with the execution levels transformer -}

-- | Interaction with AOT
instance (MonadEL m, Typeable1Monad (AOT m)) => MonadEL (AOT m) where
         current   = lift current
         up m      = AOT $ \aenv -> up (AOT.run m aenv)
         down m    = AOT $ \aenv -> down (AOT.run m aenv)
         lambda_at f l a = AOT $ \aenv -> lambda_at (\a' -> AOT.run (f a') aenv) l a

instance (Typeable1Monad m) => MonadDeploy (ELT m) where
 undeployInEnv asp@(Aspect (pc::PC (AOT (ELT m)) t1) adv hnd) aenv =
   return (deleteAsp (EAspect asp) aenv)
                   
 deployInEnv (Aspect (pc::PC (AOT (ELT m)) t1) adv hnd) aenv =
   let pcEL ldep = (PC $ return $ \jp -> do
                            lapp <- current
                            if ldep == lapp
                               then up $ runPC pc jp
                               else return False) :: PC (AOT (ELT m)) t1
       advEL ldep proceed arg = up $ adv (lambda_at proceed ldep) arg
   in do currentLevel <- current
         return (EAspect (Aspect (pcEL currentLevel) (advEL currentLevel) hnd) : aenv)

instance Typeable1Monad m => Typeable1 (ELT m) where
         typeOf1 _ = mkTyConApp (mkTyCon3 "ELT" "ELT" "ELT") [typeOf1 (undefined :: m ())]

instance MonadTrans ELT where
         lift ma = ELT $ \level -> do
                           a <- ma
                           return (a, level)

instance MonadIO m => MonadIO (ELT m) where
         liftIO = lift . liftIO

instance Monad m => MonadState s (ELT (StateT s m)) where
         get = lift get
         put s = lift (put s)

instance MonadWriter w m => MonadWriter w (ELT m) where
    tell     = lift . tell
    listen m = ELT $ \ctx -> do
                              ((a, ctx'), w) <- listen (run m ctx)
                              return ((a, w), ctx')
    pass m   = ELT $ \ctx -> pass $ do
                                   ((a, f), ctx') <- run m ctx
                                   return ((a, ctx'), f)

instance MonadCont m => MonadCont (ELT m) where
    callCC f = ELT $ \aenv ->
        callCC $ \c ->
        run (f (\a -> ELT $ \aenv' -> c (a, aenv'))) aenv

instance MonadReader r m => MonadReader r (ELT m) where
    ask       = lift ask
    local f m = ELT $ \s -> local f (run m s)