{-# LANGUAGE TemplateHaskell,
             MultiParamTypeClasses,
             FlexibleContexts,
             ScopedTypeVariables
 #-}

import AOP.Default
import AOP.Internal.TestHelper
import System.IO.Unsafe

main = runAOIDTests

runAOIDTests =
         let tests = TestList [ test_client1,
                                test_client3,
                                test_client4,
                                test_client5,
                                test_client6,
                                test_client7,
                                test_typc1,
                                test_typc2,
                                test_typc3,
                                test_typc4,
                                test_poly1,
                                test_poly2,
                                test_poly2',
                                test_poly3',
                                test_poly4,
                                test_poly6
                              ]
         in runTestTT tests
--------------------------------------------------------------------------------

cat_a :: (Typeable1Monad m) => String -> m String
cat_a s = return (s ++ "a")

adv :: (Typeable1Monad m) => Advice (Int -> m Int)
adv proceed n = proceed (n+2)

adv2 :: (Typeable1Monad m) => Advice (Int -> m Int)
adv2 proceed n = proceed 32

adv3 :: (Typeable1Monad m) => Advice (Int -> m Int)
adv3 proceed n = proceed (n-100)


-- Type of programs and runner function
type AOID a = AOT Identity a
runAOID :: AOID a -> a
runAOID c = runIdentity (runAOT c)

successor :: Typeable1Monad m => Int -> m Int
successor n = return (n+1)

successor2 :: Typeable1Monad m => Int -> AOT m Int
successor2 n = successor # n

successor3 :: Monad m => Int -> m Int
successor3 n = return (n+1)

successor'' :: Int -> AOID Int
successor'' n = return (n+1)

successor2'' :: Int -> AOID Int
successor2'' n = successor'' # n

client1 n = runAOID (program n)
program :: Int -> AOID Int
program n = let s = successor
            in do 
               deploy (aspect (pcCall s) adv)
               s # n
-- Tests
prop_client1 s = client1 s == runAOID (successor (s+2))
test_client1 = TestLabel "Client 1" (testProperty prop_client1)

-- Problem with stable names because there are nested functions. A workaround is to use let bindings.
client3 n = runAOID (program3 n)
program3 :: Int -> AOID Int
program3 n = let s  = successor''  -- :: Int -> AOID Int
                 s2 = successor2'' -- :: Int -> AOID Int
             in do 
              deploy (aspect (pcCall successor'') adv2)
              successor2'' # n
-- Tests
prop_client3 n = client3 n == 33
test_client3 = TestLabel "Client 3" (testProperty prop_client3)


-- Same problem with nested functions
client4 n = runAOID (program4 n)
program4 :: Int -> AOID Int
program4 n = let s = successor''
                 s2 = successor2''
             in do 
                deploy (aspect (pcCall successor'') adv)
                successor2'' # n
-- Tests
prop_client4 n = client4 n == runAOID (successor (n+2))
test_client4 = TestLabel "Client 4" (testProperty prop_client4)


-- Testing undeploy
client5 n = runAOID (program5 n)
program5 :: Int -> AOID Int
program5 n = let s = successor 
             in do
                x <- s # n -- x = 4
                let a = (aspect (pcCall s) (adv::Advice (Int-> AOID Int)))
                deploy a
                y <- s # n -- y = 6
                undeploy a
                z <- s # n -- z = 1
                return (x + y + z)
-- Tests
prop_client5 n = client5 n == 2 * (runAOID (successor(n))) + runAOID (successor(n+2))
test_client5 = TestLabel "Client 5" (testProperty prop_client5)


adv6 :: (Typeable1Monad m) => Advice (Int -> m a)
adv6 proceed n = proceed (n+2)

-- Testing pcType pointcut
client6 n = runAOID (program6 n)
program6 :: Int -> AOID Int
program6 n = do -- we specify the type signature using undefined :: sig. We could also define a macro in Template Haskell.
                -- the aspect applies on all functions that are more specific that the given type
                deploy (aspect (pcType (undefined :: Int -> AOID a)) adv6)
                x <- successor # n
                y <- successor2 # n
                return (x+y)
-- Tests
prop_client6 n = client6 n == runAOID (successor (n+2)) + runAOID (successor (n+4))
test_client6 = TestLabel "Client 6" (testProperty prop_client6)


client7 n = runAOID (program7 n)
program7 :: Int -> AOID Int
program7 n = let s  = successor
                 s2 = successor2
             in do -- all jps with the type, but pcCalls to successor2
                   deploy (aspect (pcType (undefined :: Int -> AOID Int) `pcAnd` pcNot (pcCall s2)) adv)
                   x <- s # n
                   y <- s2 # n
                   return (x+y)
-- Tests
prop_client7 n = client7 n == runAOID (successor (n+2)) + runAOID (successor (n+2))
test_client7 = TestLabel "Client 7" (testProperty prop_client7)







{------ Testing pcType ------}
genericAdvice :: (Typeable1Monad m) => Advice (a -> m b)
genericAdvice proceed = proceed

-- pcType: concrete / advice: concrete
typc1 n = runAOID (typprogram1 n)
typprogram1 :: Int -> AOID Int
typprogram1 n = do
                 deploy (aspect (pcType (undefined :: Int -> AOID Int)) adv)
                 x <- successor # n  -- succ(n+2)
                 y <- successor2 # n -- succ(n+4) advice triggers for successor2 and its woven nested app of successor
                 return (x+y)        -- succ(n+2) + succ(n+4)
-- Tests
prop_typc1 n = typc1 n == runAOID (successor (n+2)) + runAOID (successor (n+4))
test_typc1 = TestLabel "Type PC 1" (testProperty prop_typc1)


-- pcType: concrete / advice: polymorphic (with no constraints)
typc2 n = runAOID (typprogram2 n)
typprogram2 :: Int -> AOID Int
typprogram2 n = do
                  -- the generic advice cannot do much. We make it print something.
                  deploy (aspect (pcType (undefined :: Int -> AOID Int)) genericAdvice)
                  x <- successor # n  -- succ(n) 
                  y <- successor2 # n -- succ(n) advice triggers for successor2 and its woven nested app of successor
                  return (x+y) -- succ(n) + succ(n)
-- Tests
prop_typc2 n = typc2 n == 2 * runAOID (successor n)
test_typc2 = TestLabel "Type PC 2" (testProperty prop_typc2)

strAdvice :: (Typeable1Monad m) => Advice (String -> m b)
strAdvice proceed arg = proceed (arg ++ "_advice_")

-- pcType: polymorphic / advice: polymorphic
typc3 n = runAOID (typprogram3 n)
typprogram3 :: Int -> AOID Int
typprogram3 n = do
                 deploy (aspect (pcType (undefined :: String -> AOID b)) strAdvice)
                 z <- cat_a # "hello"   -- hello_advice_a
                 x <- successor2 # n    -- returns suc(n)
                 if (z == "hello_advice_a")
                    then return x       -- 1 implies advice didn't trigger on successor2, as expected
                    else error "Shouldn't happen"
-- Tests
prop_typc3 n = typc3 n == runAOID (successor n)
test_typc3 = TestLabel "Type PC 3" (testProperty prop_typc3)


typc4 n = runAOID (typprogram4 n)
typprogram4 :: Int -> AOID Int
typprogram4 n = do -- advice should trigger 3 times
                  deploy (aspect (pcType (undefined :: a -> AOID b)) genericAdvice)
                  z <- cat_a # "hello"
                  successor2 # n
-- Tests
prop_typc4 n = typc4 n == runAOID (successor n)
test_typc4 = TestLabel "Type PC 4" (testProperty prop_typc4)


{-------- Testing Parametric Polymorphism --------}

{- Advising a pcCall to a polymorphic function with a concrete advice.
   To do so, you must instantiate the pcCall to the polymorphic type so it 
   matches the type of the advice you want to deploy. -}
polyHead :: [a] -> AOID a
polyHead xs = return (head xs)

concreteAdvice :: Advice ([Int] -> AOID Int)
concreteAdvice proceed arg = proceed ([42] ++ arg)

-- not passing tests, maybe problem with pcCall pc
poly1 xs = runAOID (polyProgram1 xs)
polyProgram1 :: [Int] -> AOID Int
polyProgram1 xs = do 
                     deploy (aspect (pcCall (polyHead :: [Int] -> AOID Int)) concreteAdvice)
                     x <- polyHead # ([1, 2, 3] :: [Float])
                     polyHead # xs
-- Tests
prop_poly1 xs = poly1 xs == (42 :: Int)
test_poly1 = TestLabel "Polymorphism Case 1" (testProperty prop_poly1)


{- Advising a concrete function with a polymorphic advice.
   Case 1: using pcCall pc
   Case 2: using type pc
 -}
concreteFunction :: Int -> AOID Int
concreteFunction = return

concreteFunction2 :: Float -> AOID Float
concreteFunction2 = return

polyAdvice ::Advice (Int -> AOID Int)
polyAdvice proceed arg = proceed (arg+100)

-- using pcCall
poly2 n = runAOID (polyProgram2 n)
polyProgram2 :: Int -> AOID Int
polyProgram2 n = do
                   deploy (aspect (pcCall concreteFunction)  polyAdvice)
                   concreteFunction # n
-- Tests
prop_poly2 n = poly2 n == (n+100)
test_poly2 = TestLabel "Polymorphism Case 2" (testProperty prop_poly2)

-- using pcType
-- This more general signature works OK with pcType
polyAdvice' :: (Typeable1Monad m, Num a, Show a) => Advice (a -> (AOT m) b)
polyAdvice' proceed arg = proceed (arg+100)

poly2' n = runAOID (polyProgram2' n)
polyProgram2' :: Int -> AOID Int
polyProgram2' n = do
                   deploy (aspect (pcType (undefined :: Int -> AOID Int)) polyAdvice')
                   concreteFunction # n -- n+1
-- Tests
prop_poly2' n = poly2' n == (n+100)
test_poly2' = TestLabel "Polymorphism Case 2' " (testProperty prop_poly2')


-- using pcType
poly3' x = runAOID (polyProgram3' x)
polyProgram3' :: Float -> AOID Float
polyProgram3' x = do deploy (aspect (pcType (undefined :: Float -> AOID Float)) polyAdvice')
                     concreteFunction2 # x                    
-- Tests
-- Since we are using floating point we establish a tolerance
prop_poly3' x = abs (poly3' x - (x+100)) < 0.0001 where types = x :: Float
test_poly3' = TestLabel "Polymorphism Case 3' " (testProperty prop_poly3')


polyAdvice2 :: (Num a, Show a, Show b) => Advice (a -> AOID b)
polyAdvice2 proceed arg = proceed (arg +100)

constraintPoly x = runAOID (constraintPolyProgram x)
constraintPolyProgram :: Float -> AOID Float
constraintPolyProgram x = do deploy (aspect (pcCall concreteFunction2) polyAdvice2)
                             y <- concreteFunction2 # x
                             return y
-- Tests
prop_constraintPoly x = abs(constraintPoly x - (100+x)) < 0.0001 where types = x :: Float
test_constraintPoly = TestLabel "Constrained Polymorphic Advice" (testProperty prop_constraintPoly)

concreteFunction3 :: Int -> AOID String
concreteFunction3 n = return (show n) 

poly4 n = runAOID (polyProgram4 n)
polyProgram4 :: Int -> AOID Int
polyProgram4 n = do
                   deploy (aspect (pcCall concreteFunction)  polyAdvice)                   
                   concreteFunction # n
-- Tests
prop_poly4 n = poly4 n == (n+100)
test_poly4 = TestLabel "Polymorphism Case 4" (testProperty prop_poly4)


{- Advising a pcCall to a polymorphic function with an also polymorphic advice.
   As usual, the type of the advice cannot be more specific than the type of the pointcut.

   Case 1: Using pcCall pc
   Case 2: Using type pc

   If the polymorphic target function has no constraints. It works OK. Else, see TestMacros.
 -}

polyHead2 :: (Typeable1Monad m, Num a) => [a] -> m a
polyHead2 xs = return (head xs)

polyAdvice3 :: (Typeable1Monad m, Num a) => Advice ([a] -> m a)
polyAdvice3 proceed arg = proceed ([47] ++ arg)

polyAdvice3' :: (Typeable1Monad m) => Advice ([a] -> (AOT m) a)
polyAdvice3' proceed arg = proceed (reverse arg)

-- Using pcType
-- If we omit the annotations the advice is not triggered.
poly5' :: [Int] -> Int
poly5' xs = runAOID (polyProgram5' xs)
polyProgram5' :: [Int] -> AOID Int
polyProgram5' xs = do
                     deploy (aspect (pcType (undefined :: [a] -> AOID a)) polyAdvice3')
                     polyHead2 # xs
-- Tests
prop_poly5' xs = not (null xs) ==> poly5' xs == head (reverse xs) where types = (xs :: [Int])
test_poly5' = TestLabel "Polymorphism Case 5'" (testProperty prop_poly5')


-- When using pcCall we need a let binding
-- One for each specific type to be used
poly6 xs = runAOID (polyProgram6 xs)
polyProgram6 xs = let f = polyHead2 :: [Int] -> AOID Int
                      g = polyHead2 :: [Float] -> AOID Float
                  in do             
                      deploy (aspect (pcCall f) polyAdvice3)
                      deploy (aspect (pcCall g) polyAdvice3)
                      x <- f # ([1, 2, 3] :: [Int])                      
                      y <- g # ([1, 2, 3] :: [Float])                     
                      z <- f # xs
                      -- Can't do wrong applications
                      -- w <- polyHead2 # ['a', 'b', 'c']
                      return (x + (round y) + z) -- 47 * 3 = 141
-- Tests
prop_poly6 xs = (xs /= []) ==> poly6 xs == 141
test_poly6 = TestLabel "Polymorphism Case 6" (testProperty prop_poly6)