{-# LANGUAGE MultiParamTypeClasses,
             ScopedTypeVariables,
             RankNTypes,
             FlexibleContexts,
             FlexibleInstances,
             ExistentialQuantification,
             IncoherentInstances
 #-}

import AOP.Default
import AOP.Internal.TestHelper

-- Monadic stack (and run function) for possibly-interfering state
-- type AOState = NIAOT (StateT Int) Identity 

-- runAOState c n = runIdentity (evalStateT (runNIAOT c) n)

-- Monadic stack (and run function) for non-interfering state
type NIAOState = NIAOT (StateT Bool) (StateT Int Identity)

runNIAOState c nbase nadv = runIdentity (evalStateT (evalStateT (runNIAOT c) nadv) nbase)

-- A tagged caching function
-- it increases the current cache by m+1
cachingSucc :: Function Int (NIAOState Int)
cachingSucc = mkFunction (\ m -> do n <- niLift (lift get)
                                    (niLift . lift . put) (n+m+1)
                                    return (n+m)) cachTag

-- Monadic successor function
successor :: Function Int (NIAOState Int)
successor = Function (\ n -> return (n+1)) succTag

pureAdv :: Monad m => Advice (Int -> m Int)
pureAdv proceed n = proceed (n+10)


-- Fresh tags for the cachingSucc and succTag functions
-- To be used in our notion of equality based.
cachTag = newTag
succTag = newTag

-- initializing the Int and Bool states on the monadic stack
-- the integer state is only available to aspects, and the boolean state only to base code
clientNI s = runNIAOState (programNI s) 1 False

-- the pointcut sequence triggers execution of pureAdv
-- the cachingSucc correctly updates the state without interfering with the base state
programNI :: Int -> NIAOState Int
programNI s = do deploy (niAspect (pcCall successor `pcSeq` pcCall cachingSucc) pureAdv)
                 cachingSucc # 1
                 successor # 1
                 cachingSucc # s

prop_clientNI s = clientNI s == 13+s
test_clientNI = TestLabel "Client NI" (testProperty prop_clientNI)

main = let tests = TestList [test_clientNI]
       in runTestTT tests
